(**
 ** $Id: ImageColor.sml 95 2005-12-16 18:30:44Z rlpm $
 **
 ** An image processing library for Standard ML
 ** Author: Rory McGuire <rlpm at cs dot unm dot edu>
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License as
 ** published by the Free Software Foundation; either version 2 of the
 ** License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ** General Public License for more details.
 **
 ** You can obtain a copy of the GPL at:
 **   http://www.fsf.org/copyleft/gpl.html.
 **
 **)

structure ImageColor :> IMAGE_COLOR =
struct
  open Image_Image
  type image = (real * real * real) image

  fun fromGrey g =
      map (fn x => (x,x,x)) g

  val red = map (fn (r:real,_:real,_:real) => r)
  val green = map (fn (_:real,g:real,_:real) => g)
  val blue = map (fn (_:real,_:real,b:real) => b)

  fun toRGB img =
      {red=red img,green=green img, blue=blue img}

  fun fromRGB {red,green,blue} =
      if nRows red = nRows green andalso nRows red = nRows blue andalso
         nCols red = nCols green andalso nCols red = nCols blue
      then tabulate
               (nRows red, nCols red,
                (fn (i,j) => (sub(red,i,j),
                              sub(green,i,j),
                              sub(blue,i,j))))
      else raise Fail "fromRGB: Images are of unequal sizes"

  fun fromGreyHot img =
      let
          val oneThird = Real./(1.0,3.0)
          val twoThirds = Real./(2.0,3.0)
      in
          tabulate(nRows img, nCols img,
                   (fn (i,j) =>
                       let
                           val v = sub(img,i,j)
                       in
                           if v <= oneThird
                           then (Real.*(3.0,v),0.0,0.0)
                           else if v <= twoThirds
                           then (1.0,Real.*(3.0,Real.-(v,oneThird)),0.0)
                           else (1.0,1.0,Real.*(3.0,Real.-(v,twoThirds)))
                       end))
      end

  fun normalize (min,max) =
      let
          val range = max - min
          val zero = Real.==(0.0,range)
      in
          mapi (fn (i,j,(r,g,b)) =>
                   if r < min orelse r > max orelse
                      g < min orelse g > max orelse
                      b < min orelse b > max
                   then raise Fail ("Value (" ^
                                    Real.toString r ^ "," ^
                                    Real.toString g ^ "," ^
                                    Real.toString b ^ ") at (" ^
                                    Int.toString i ^ "," ^
                                    Int.toString j ^
                                    ") out of range")
                   else if zero
                   then (0.0,0.0,0.0)
                   else
                       let
                           fun f x = (x - min) / (max - min)
                       in
                           (f r, f g, f b)
                       end)
      end

  fun maximizeContrast img = 
      normalize
          (fold (fn ((r,g,b),(min,max)) =>
                    (Real.min(min,Real.min(r,Real.min(g,b))),
                     Real.max(max,Real.max(r,Real.max(g,b)))))
                (let
                     val (r,g,b) = sub (img,0,0)
                 in
                     (Real.min(r,Real.min(g,b)),
                      Real.max(r,Real.max(g,b)))
                 end)
                img)
          img

  fun norm256 img =
      Array2.tabulate
          Array2.RowMajor
          (nRows img, nCols img,
           (fn (i,j) =>
               let
                   val (r,g,b) = sub (img,i,j)
               in
                   if r < 0.0 orelse r > 1.0 orelse
                      g < 0.0 orelse g > 1.0 orelse
                      b < 0.0 orelse b > 1.0
                   then raise Fail ("Value (" ^
                                Real.toString r ^ "," ^
                                Real.toString g ^ "," ^
                                Real.toString b ^ ") at (" ^
                                Int.toString i ^ "," ^
                                Int.toString j ^
                                ") out of range")
                   else (Real.round (r * 255.0),
                         Real.round (g * 255.0),
                         Real.round (b * 255.0))
               end))

  fun writeFile (p,f) (img,filename) = 
      let
          val array = norm256 img
          val file = TextIO.openOut filename
          val header = p ^ "\n" ^
                       "# CREATOR: Image Processing Library for StandardML " ^
                       "$Revision: 95 $\n" ^
                       Int.toString (nCols img) ^ " " ^
                       Int.toString (nRows img) ^ "\n" ^
                       "255\n"
          val _ = TextIO.output(file,header)
          fun output x = TextIO.output(file,f x)
          val _ = Array2.fold
                      Array2.RowMajor
                      (fn ((r,g,b),_) => (output r, output g, output b))
                      ((),(),())
                      array
      in
          TextIO.closeOut file
      end

  val writePPMA = writeFile ("P3",fn x=>" "^Int.toString x)
  val writePPM = writeFile ("P6",String.str o Char.chr)

  fun readPPM filename = 
      let
          val file = BinIO.openIn filename
          fun input1 file = case BinIO.input1 file of
                                NONE => NONE
                              | SOME c => (SOME o Char.chr o Word8.toInt) c
          fun eatLine () = if input1 file = SOME #"\n" then () else eatLine ()
          fun readRec acc =
              case input1 file of
                  NONE => acc
                | SOME(x) => if x = #"#"
                             then (eatLine() ; readRec acc)
                             else if Char.isSpace x
                             then if acc = []
                                  then readRec acc 
                                  else acc
                             else readRec (x::acc)
          fun read () = (String.implode o rev o readRec) []
          fun cleanup () = (BinIO.closeIn file ;
                            raise Fail ("File format error in: " ^ filename))
          (* magic number *)
          val p = read ()
          val _ = if p = "P3" orelse p = "P6"
                  then ()
                  else cleanup ()
          val cols = case Int.fromString (read ()) of
                         NONE => cleanup ()
                       | SOME(x) => x
          val rows = case Int.fromString (read ()) of
                         NONE => cleanup ()
                       | SOME(x) => x
          val max = case Int.fromString (read ()) of
                         NONE => cleanup ()
                       | SOME(x) => Real.fromInt x
          fun readInt _ = case (if p = "P3"
                                then Int.fromString (read ())
                                else case input1 file of
                                         NONE => NONE
                                       | SOME x => SOME(Char.ord x)) of
                               NONE => cleanup ()
                             | SOME(x) => Real.fromInt x / max
          val img = tabulate
                    (rows,cols,
                     (fn _ => (readInt (), readInt (), readInt ())))
          val _ = BinIO.closeIn file
      in
          img
      end

      fun HSVtoRGB (h:real,s:real,v:real) =
          if Real.==(s,0.0)
          then (0.0,0.0,0.0)
          else
              let
                  val var_h = Real.*(h,6.0)
                  val var_i = Real.fromInt (Real.floor var_h)
                  val var_1 = Real.*(v,Real.-(1.0,s))
                  val var_2 = Real.*(v,
                                     Real.-(1.0,Real.*(s,Real.-(var_h,var_i))))
                  val var_3 = Real.*(v,
                                     Real.-(1.0,
                                            Real.*(s,
                                                   Real.-(1.0,
                                                          Real.-(var_h,
                                                                var_i)))))
              in
                  if Real.==(var_i,0.0) then (v,var_3,var_1)
                  else if Real.==(var_i,1.0) then (var_2, v, var_1)
                  else if Real.==(var_i,2.0) then (var_1, v, var_3)
                  else if Real.==(var_i,3.0) then (var_1, var_2, v)
                  else if Real.==(var_i,4.0) then (var_3, var_1, v)
                  else (v, var_1, var_2)
              end

      fun fromHSV {hue,saturation,value} =
          if nRows hue = nRows saturation andalso
             nRows hue = nRows value andalso
             nCols hue = nCols saturation andalso
             nCols hue = nCols value
          then tabulate(nRows hue,nCols hue,
                     fn (i,j) => HSVtoRGB(sub(hue,i,j),
                                          sub(saturation,i,j),
                                          sub(value,i,j)))
          else raise Fail "Image.Color.fromHSV: images not of same size"

      fun RGBtoHSV (r,g,b) =
          let
              val var_min = Real.min(Real.min(r,g),b)
              val var_max = Real.max(Real.max(r,g),b)
              val del_max = Real.-(var_max,var_min)
          in
              if Real.==(del_max,0.0)
              then (0.0,0.0,var_max)
              else 
                  let
                      val s = Real./(del_max,var_max)
                      fun del x =
                          Real./(Real.+(Real./(Real.-(var_max,x),6.0),
                                        Real./(del_max,2.0)),
                                 del_max)
                      val del_r = del r
                      val del_g = del g
                      val del_b = del b
                      val h = if Real.==(r,var_max)
                              then Real.-(del_b, del_g)
                              else if Real.==(g,var_max)
                              then Real.-(Real.+(Real./(1.0,3.0),del_r),
                                          del_b)
                              else Real.-(Real.+(Real./(2.0,3.0),del_g),
                                          del_r)
                      val h = if Real.<(h,0.0)
                              then Real.+(h,1.0)
                              else if Real.>(h,1.0)
                              then Real.-(h,1.0)
                              else h
                  in
                      (h,s,var_max)
                  end
          end

      fun toHSV img =
          let
              val img = map RGBtoHSV img
          in
              {hue=map #1 img,
               saturation=map #2 img,
               value=map #3 img}
          end
end
