(**
 ** $Id: ImageComplex.sml 92 2005-12-16 17:34:14Z rlpm $
 **
 ** An image processing library for Standard ML
 ** Author: Rory McGuire <rlpm at cs dot unm dot edu>
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License as
 ** published by the Free Software Foundation; either version 2 of the
 ** License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ** General Public License for more details.
 **
 ** You can obtain a copy of the GPL at:
 **   http://www.fsf.org/copyleft/gpl.html.
 **)

structure ImageComplex :> IMAGE_COMPLEX =
struct
  open Image_Image

  type image = Complex.complex image

  val fromGrey = map Complex.fromReal

  fun fromRect (x,y) =
      if nRows x = nRows y andalso nCols x = nCols y
      then tabulate(nRows x, nCols x,
                    (fn (i,j) => {real=sub(x,i,j),imag=sub(y,i,j)}))
      else raise Fail "fromRect: Image sizes don't match"

  val realPart : image -> ImageGrey.image = map #real
  val imagPart : image -> ImageGrey.image = map #imag

  fun toRect (img:image) =
      (realPart img, imagPart img)

  fun fromPolar (magnitude,phase) =
      if nRows magnitude = nRows phase andalso nCols magnitude = nCols phase
      then tabulate(nRows magnitude, nCols magnitude,
                    (fn (i,j) =>
                        Complex.fromPolar {magnitude=sub(magnitude,i,j),
                                           phase=sub(phase,i,j)}))
      else raise Fail "fromPolar: Image sizes don't match"

  val magnitude = map Complex.magnitude
  val phase = map Complex.phase

  fun toPolar img =
      (magnitude img, phase img)

  fun toColor img =
      ImageColor.fromHSV{value=
                             let
                                 val mag = magnitude img
                             in
                                 ImageGrey.normalize
                                     (0.0,fold Real.max 0.0 mag)
                                     (magnitude img)
                             end,
                         hue = ImageGrey.normalize
                                   (Real.~(Math.pi),Math.pi)
                                   (phase img),
                         saturation = image(nRows img, nCols img, 1.0)}

  val op+ = binop Complex.+
  val op- = binop Complex.+
  val op* = binop Complex.+
  val op/ = binop Complex.+

  fun product (one,two) =
      let
          val _ = if nCols one = nRows two then () else raise Subscript
      in
          tabulate(nRows one, nCols two,
                   (fn (i,j) =>
                       foldl Complex.+ {real=0.0,imag=0.0}
                             (List.map Complex.* 
                                       (ListPair.zip
                                            (Vector.foldl op:: []
                                                          (Image_Image.row
                                                               (one,i)),
                                             Vector.foldl op:: []
                                                          (Image_Image.column
                                                               (two,j)))))))
      end

  fun shrink (img,b) = 
      let
          val b = if b < 0.0
                  then raise Fail ("Cannot shrink by negative value " ^
                                   Real.toString b)
                  else b
      in
          map (fn x => 
                  let
                      val mag = Complex.magnitude x
                      val phase = Complex.phase x
                  in
                      if mag <= b
                      then {real=0.0,imag=0.0}
                      else Complex.fromPolar{magnitude=Real.-(mag,b),
                                             phase=phase}
                  end)
              img
      end

  structure conv = Convolve (struct
                             type a = Complex.complex
                             val zero = {real=0.0,imag=0.0}
                             val op+ = Complex.+
                             val op* = Complex.*
                             end)

  fun convolveRows knl img =
      let
          val vec = Vector.tabulate
                    (nRows img,
                     (fn i => conv.oneDim knl (row (img,i))))
      in
          mapi (fn (i,j,_) => Vector.sub(Vector.sub(vec,i),j)) img
      end

  fun convolveCols knl =
      transpose o convolveRows knl o transpose

  val convolve = conv.twoDim

  val dft = Fourier.dft2d
  val idft = Fourier.idft2d
  val fft = Fourier.fft2d
  val ifft = Fourier.ifft2d
end
