(**
 ** $Id: ImageGrey.sml 91 2005-12-16 17:15:50Z rlpm $
 **
 ** An image processing library for Standard ML
 ** Author: Rory McGuire <rlpm at cs dot unm dot edu>
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License as
 ** published by the Free Software Foundation; either version 2 of the
 ** License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ** General Public License for more details.
 **
 ** You can obtain a copy of the GPL at:
 **   http://www.fsf.org/copyleft/gpl.html.
 **)

structure ImageGrey :> IMAGE_GREY =
struct
  open Image_Image

  type image = real image

  fun normalize (min,max) =
      let
          val range = max - min
          val zero = Real.==(0.0,range)
      in
          mapi (fn (i,j,x) =>
                   if x < min orelse x > max
                   then raise Fail ("Value " ^ Real.toString x ^
                                    " at (" ^ Int.toString i ^
                                    "," ^ Int.toString j ^
                                    ") out of range")
                   else if zero
                   then 0.0
                   else (x - min) / range)
      end

  fun maximizeContrast img = 
      normalize (fold (fn (x,(min,max)) => (if x < min then x else min,
                                            if x > max then x else max))
                      (sub (img,0,0), sub (img,0,0)) img) img

  fun norm256 img =
      Array2.tabulate
          Array2.RowMajor
          (nRows img, nCols img,
           (fn (i,j) =>
               let
                   val x = sub (img,i,j)
               in
                   if x < 0.0 orelse x > 1.0
                   then raise Fail ("Value " ^
                                    Real.toString x ^
                                    " at (" ^
                                    Int.toString i ^
                                    "," ^
                                    Int.toString j ^
                                    ") out of range")
                   else Real.round (x * 255.0)
               end))

  fun writeFile (p,f) (img,filename) =
      let
          val array = norm256 img
          val file = TextIO.openOut filename
          val header = p ^ "\n" ^
                       "# CREATOR: Image Processing Library for StandardML " ^
                       "$Revision: 91 $\n" ^
                       Int.toString (nCols img) ^ " " ^
                       Int.toString (nRows img) ^ "\n" ^
                       "255\n"
          val _ = TextIO.output(file,header)
          val _ = Array2.fold
                      Array2.RowMajor
                      (fn (x,_) => TextIO.output(file,f x))
                      ()
                      array
      in
          TextIO.closeOut file
      end

  val writePGMA = writeFile ("P2",fn x=>" "^Int.toString x)
  val writePGM = writeFile ("P5",String.str o Char.chr)

  fun readPGM filename = 
      let
          val file = BinIO.openIn filename
          fun input1 file = case BinIO.input1 file of
                                NONE => NONE
                              | SOME c => (SOME o Char.chr o Word8.toInt) c
          fun eatLine () = if input1 file = SOME #"\n" then () else eatLine ()
          fun readRec acc =
              case input1 file of
                  NONE => acc
                | SOME(x) => if x = #"#"
                             then (eatLine() ; readRec acc)
                             else if Char.isSpace x
                             then if acc = []
                                  then readRec acc 
                                  else acc
                             else readRec (x::acc)
          fun read () = (String.implode o rev o readRec) []
          fun cleanup () = (BinIO.closeIn file ;
                            raise Fail ("File format error in: " ^ filename))
          (* magic number *)
          val p = read ()
          val _ = if p = "P2" orelse p = "P5"
                  then ()
                  else cleanup ()
          val cols = case Int.fromString (read ()) of
                         NONE => cleanup ()
                       | SOME(x) => x
          val rows = case Int.fromString (read ()) of
                         NONE => cleanup ()
                       | SOME(x) => x
          val max = case Int.fromString (read ()) of
                         NONE => cleanup ()
                       | SOME(x) => Real.fromInt x
          val img = tabulate
                    (rows,cols,
                     (fn (i,j) => case (if p = "P2"
                                        then Int.fromString (read ())
                                        else case input1 file of
                                                 NONE => NONE
                                               | SOME x => SOME(Char.ord x)) of
                                      NONE => cleanup ()
                                    | SOME(x) => Real.fromInt x / max))
          val _ = BinIO.closeIn file
      in
          img
      end

  val op+ = binop Real.+
  val op- = binop Real.-
  val op* = binop Real.*
  val op/ = binop Real./

  fun product (one,two) =
      let
          val _ = if nCols one = nRows two then () else raise Subscript
      in
          tabulate(nRows one, nCols two,
                   (fn (i,j) =>
                       foldl Real.+ 0.0
                             (List.map Real.* 
                                       (ListPair.zip
                                            (Vector.foldl op:: []
                                                          (Image_Image.row
                                                               (one,i)),
                                             Vector.foldl op:: []
                                                          (Image_Image.column
                                                               (two,j)))))))
      end

  fun histogram img =
      let
          val img = norm256 img
          val v = Array.array (256,0)
          val _ = Array2.fold
                      Array2.RowMajor
                      (fn (x,_) =>
                           Array.update(v,x,Int.+(Array.sub(v,x),1)))
                      ()
                      img
      in
          Array.foldr op:: [] v
      end

  fun jointHistogram (one,two) =
      let
          val _ = if nRows one = nRows two andalso nCols one = nCols two
                  then ()
                  else raise Subscript
          val one = norm256 one
          val two = norm256 two
          val v = Array2.array (256,256,0)
          val _ = Array2.foldi
                      Array2.RowMajor
                      (fn (i,j,x,_) =>
                          let
                              val y = Array2.sub(two,i,j)
                          in
                              Array2.update(v,x,y,Int.+(Array2.sub(v,x,y),1))
                          end)
                      ()
                      {base=one, row=0, col=0, nrows=NONE, ncols=NONE}
      in
          v
      end

  fun shrink (img,b) = 
      let
          val b = if b < 0.0
                  then raise Fail ("Cannot shrink by negative value " ^
                                   Real.toString b)
                  else b
      in
          map (fn x => if Real.abs x < b
                         then 0.0
                         else if x < 0.0
                         then Real.+(x,b)
                         else Real.-(x,b))
              img
      end

  structure conv = Convolve (struct
                             type a = Real.real
                             val zero = 0.0
                             val op+ = Real.+
                             val op* = Real.*
                             end)

  fun convolveRows knl img =
      let
          val vec = Vector.tabulate
                    (nRows img,
                     (fn i => conv.oneDim knl (row (img,i))))
      in
          mapi (fn (i,j,_) => Vector.sub(Vector.sub(vec,i),j)) img
      end

  fun convolveCols knl =
      transpose o convolveRows knl o transpose

  val convolve = conv.twoDim

end
