(**
 ** $Id: Image_Image.sml 127 2006-02-14 18:44:00Z rlpm $
 **
 ** An image processing library for Standard ML
 ** Author: Rory McGuire <rlpm at cs dot unm dot edu>
 **
 ** This program is free software; you can redistribute it and/or
 ** modify it under the terms of the GNU General Public License as
 ** published by the Free Software Foundation; either version 2 of the
 ** License, or (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 ** General Public License for more details.
 **
 ** You can obtain a copy of the GPL at:
 **   http://www.fsf.org/copyleft/gpl.html.
 **)

structure Image_Image :> IMAGE_IMAGE =
struct
  type 'a image = 'a Array2.array

  val image = Array2.array

  fun fromList ls =
      Array2.fromList ls handle Size => raise Size

  fun toList img =
      List.tabulate
          (Array2.nRows img,
           (fn x => Vector.foldr op:: [] (Array2.row (img,x))))

  fun tabulate x = Array2.tabulate Array2.RowMajor x
  fun tabulateFilter (n,m,f) =
      let
          val n_2 = n div 2
          val m_2 = m div 2
      in
          tabulate (n,m,(fn (i,j) => f (if i < n_2 then i else i - n,
                                        if j < m_2 then j else j - m)))
      end

  val sub = Array2.sub

  val nRows = Array2.nRows
  val nCols = Array2.nCols

  val row = Array2.row
  val column = Array2.column

  fun mapi f img = tabulate
                       (nRows img,
                        nCols img,
                        (fn (i,j) => f (i,j,(sub (img,i,j)))))

  fun map f = mapi (fn (_,_,x) => f x)

  fun fold f = Array2.fold Array2.RowMajor f

  fun foldi f init img = Array2.foldi
                             Array2.RowMajor 
                             f
                             init
                             {base=img, row=0, col=0, nrows=NONE, ncols=NONE}

  fun crop (img,i0,j0,n,m) =
      tabulate(n,m,(fn(i,j) => sub(img,i0+i,j0+j)))

  fun pad a (img,n,m) =
      let
          val r = nRows img
          val c = nCols img
      in
          tabulate(n,m,
                   (fn(i,j) => if i<r andalso j<c
                               then sub(img,i,j)
                               else a))
      end

  fun binop f (one,two) =
      let
          val _ = if nCols one = nCols two andalso
                     nRows one = nRows two
                  then ()
                  else raise Subscript
      in
          tabulate (nRows one, nCols one,
                    (fn(i,j)=>f(sub(one,i,j),sub(two,i,j))))
      end

  fun transpose img = 
      tabulate (nCols img, nRows img, (fn (i,j) => sub(img,j,i)))

  fun flip img =
      tabulate (nRows img, nCols img,
                (fn (i,j) => sub(img,nRows img - i - 1, j)))

  fun flop img =
      tabulate (nRows img, nCols img,
                (fn (i,j) => sub(img, i, nCols img - j - 1)))

  fun rotateCW img =
      tabulate (nCols img, nRows img,
                (fn (i,j) => sub(img,nRows img - j - 1, i)))

  fun rotateCCW img =
      tabulate (nCols img, nRows img,
                (fn (i,j) => sub(img,j,nCols img - i - 1)))

  fun downSampleRows img =
      tabulate ((nRows img div 2) + (nRows img mod 2), nCols img,
                (fn (i,j) => sub(img,i*2,j)))

  fun downSampleCols img =
      tabulate (nRows img, (nCols img div 2) + (nCols img mod 2),
                (fn (i,j) => sub(img,i,j*2)))

  fun downSample img =
      tabulate ((nRows img div 2) + (nRows img mod 2),
                (nCols img div 2) + (nCols img mod 2),
                (fn (i,j) => sub(img,i*2,j*2)))

  fun upSampleRows a img =
      tabulate (nRows img * 2, nCols img,
                (fn (i,j) => if i mod 2 = 0
                             then sub(img,i div 2,j)
                             else a))

  fun upSampleCols a img =
      tabulate (nRows img, nCols img * 2,
                (fn (i,j) => if j mod 2 = 0
                             then sub(img,i,j div 2)
                             else a))

  fun upSample a img =
      tabulate (nRows img * 2, nCols img * 2,
                (fn (i,j) => if i mod 2 = 0 andalso j mod 2 = 0
                             then sub(img,i div 2,j div 2)
                             else a))

  fun leftToRight ls =
      let
          val def = sub(hd ls,0,0)
          val rows = foldl Int.max 0 (List.map nRows ls)
          val ls = List.map (fn x => (nCols x,x)) ls
          val (cols,ls) = List.foldl
                           (fn ((i,x),(tot,acc)) =>
                               (Int.+(i,tot),(tot,x)::acc))
                           (0,[])
                           ls
          val ls = rev ls
          val ret = image(rows,cols,def)
          val _ = foldl
                  (fn ((offset,img),_) =>
                      Array2.foldi
                          Array2.RowMajor
                          (fn (i,j,a,_) =>
                               Array2.update(ret,i,Int.+(offset,j),a))
                          ()
                          {base=img,row=0,col=0,nrows=NONE,ncols=NONE})
                  ()
                  ls
      in
          ret
      end

  fun topToBottom ls =
      (transpose o leftToRight o List.map transpose) ls
end
